#!/usr/bin/env python

"""
Moin wiki parsing functionality.

Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict
from moinformat.tree.moin import Block, Region, Text
import re

# Pattern management.

ws_excl_nl = r"[ \f\r\t\v]"
quotes = "['" '"]'              # ['"]
dotall = r"(.|\n)"

def choice(l):

    "Return a pattern matching a choice of patterns in 'l'."

    return "(%s)" % "|".join(l)

def excl(s):

    "Return a non-matching pattern for 's'."

    return "(?!%s)" % s

def expect(s):

    "Return a pattern expecting 's'."

    return "(?=%s)" % s

def group(name, s):

    "Return a pattern group having 'name' and the pattern string 's'."

    return "(?P<%s>%s)" % (name, s)

def optional(s):

    "Return an optional pattern."

    return "(?:%s)?" % s

def recur(name):

    "Return a test for a recurrence of group 'name'."

    return "(?P=%s)" % name

def repeat(s, min=None, max=None):

    "Return a pattern matching 's' for the given 'min' and 'max' limits."

    return "%s{%s,%s}" % (s, min is not None and min or "",
                             max is not None and max or "")

def get_pattern(s):

    "Return a compiled regular expression for the given pattern 's'."

    return re.compile(s, re.UNICODE | re.MULTILINE)

def get_patterns(syntax):

    """
    Define patterns for the regular expressions in the 'syntax' mapping. In each
    pattern, replace \N with a pattern for matching whitespace excluding
    newlines.
    """

    patterns = {}
    for name, value in syntax.items():
        value = value.replace(r"\N", ws_excl_nl)
        value = value.replace(r"\Q", quotes)
        value = value.replace(r"\E", dotall)
        patterns[name] = get_pattern(value)
    return patterns

def get_subset(d, keys):

    "Return a subset of 'd' having the given 'keys'."

    subset = {}
    for key in keys:
        subset[key] = d[key]
    return subset



# Tokenising functions.

class TokenStream:

    "A stream of tokens taken from a string."

    def __init__(self, s, pos=0):
        self.s = s
        self.pos = pos

        # Match details.

        self.match = None
        self.queued = None
        self.match_start = None

        # Pattern name details.

        self.matching = None

    def rewind(self, length):

        "Rewind in the string by 'length'."

        self.pos -= min(length, self.pos)

    def queue_match(self):

        "Rewind in the string to the start of the last match."

        self.queued = self.match

    def read_until(self, patterns, remaining=True):

        """
        Find the first match for the given 'patterns'. Return the text preceding
        any match, the remaining text if no match was found, or None if no match
        was found and 'remaining' is given as a false value.
        """

        if self.queued:
            self.match = self.queued
            self.queued = None
        else:
            self.match_start = None
            self.matching = None

            # Find the first matching pattern.

            for pattern_name, pattern in patterns.items():
                match = pattern.search(self.s, self.pos)
                if match:
                    start, end = match.span()
                    if self.matching is None or start < self.start:
                        self.start = start
                        self.matching = pattern_name
                        self.match = match

        if self.matching is None:
            if remaining:
                return self.s[self.pos:]
            else:
                return None
        else:
            return self.s[self.pos:self.start]

    def match_group(self, group=1):

        """
        Return the matched text, updating the position in the stream. If 'group'
        is specified, the indicated group in a match will be returned.
        Typically, group 1 should contain all pertinent data, but groups defined
        within group 1 can provide sections of the data.
        """

        self.update_pos()

        if self.match:
            try:
                return self.match.group(group)
            except IndexError:
                return ""
        else:
            return None

    def match_groups(self, groups=None):

        "Return the match 'groups', or all groups if unspecified."

        self.update_pos()

        if self.match:
            if groups is None:
                return self.match.groups()
            else:
                return self.match.groups(groups)
        else:
            return []

    def update_pos(self):

        "Update the position in the stream."

        if self.match:
            _start, self.pos = self.match.span()
        else:
            self.pos = len(self.s)



# Parser abstractions.

class ParserBase:

    "Common parsing methods."

    region_pattern_names = None

    def __init__(self, metadata, parsers=None, root=None):

        """
        Initialise the parser with the given 'metadata' and optional 'parsers'.
        An optional 'root' indicates the document-level parser.
        """

        self.metadata = metadata
        self.parsers = parsers
        self.root = root

    def get_parser(self, format_type):

        """
        Return a parser for 'format_type' or None if no suitable parser is found.
        """

        cls = self.parsers and self.parsers.get(format_type)
        if cls:
            return cls(self.metadata, self.parsers, self.root or self)
        else:
            return None

    def get_patterns(self, pattern_names):

        "Return a mapping of the given 'pattern_names' to patterns."

        return get_subset(self.patterns, pattern_names)

    def get_items(self, s, pos=0):

        "Return a sequence of token items for 's' and 'pos'."

        return TokenStream(s, pos)

    def set_region(self, items, region):

        "Set the 'items' used to populate the given 'region'."

        self.items = items
        self.region = region

    def read_until(self, pattern_names, remaining=True):

        """
        Read the next portion of input, matching using 'pattern_names'. Return
        the text preceding any match, the remaining text if no match was found,
        or None if no match was found and 'remaining' is given as a false value.
        """

        return self.items.read_until(self.get_patterns(pattern_names))

    def match_group(self, group=1):

        """
        Return the group of the matching pattern with the given 'group' number.
        """

        return self.items.match_group(group)

    def matching_pattern(self):

        "Return the name of the matching pattern."

        return self.items.matching

    def match_groups(self):

        "Return the number of groups in the match."

        return self.items.match_groups()

    # Parser methods invoked from other objects.

    def parse(self, s):

        """
        Parse page text 's'. Pages consist of regions delimited by markers.
        """

        self.items = self.get_items(s)
        self.region = self.parse_region()
        return self.region

    def parse_region_content(self, items, region):

        "Parse the data provided by 'items' to populate a 'region'."

        self.set_region(items, region)

        # Parse inline and opaque regions.

        if not region.transparent:
            pattern_names = ["regionend"]

        # Define a block to hold text.

        else:
            self.new_block(region)
            pattern_names = self.region_pattern_names

        # Start parsing.

        if pattern_names:
            self.parse_region_details(region, pattern_names)

        # Reset the type if the region was not inline.

        if region.type == "inline":
            first = region.nodes and region.nodes[0]
            if first and isinstance(first, Text) and first.multiline():
                region.type = None

    # Top-level parser handler methods.

    def parse_region(self, level=0, indent=0, type=None):

        """
        Parse the data to populate a region with the given 'level' at the given
        'indent' having the given initial 'type'.
        """

        region = Region([], level, indent, type)

        # Parse section headers and directives, then parse according to region
        # type.

        self.parse_region_header(region)
        self.parse_region_directives(region)
        self.parse_region_type(region)

        return region

    def parse_region_type(self, region):

        """
        Use configured parsers to parse 'region' based on its type.
        """

        # Find an appropriate parser given the type.

        parser = self.get_parser(region.type)
        if not parser:
            region.transparent = False
        parser = parser or self.get_parser("moin")
        parser.parse_region_content(self.items, region)

    def parse_region_header(self, region):

        """
        Parse the region header, setting it on the 'region' object.
        """

        if self.read_until(["header"], False) == "": # None means no header
            region.args = self.match_group("args")
            region.type = region.args.split(" ", 1)[0]

    def parse_region_directives(self, region):

        """
        Parse any directives immediately after the region header, adding them to
        the 'region' object.
        """

        while True:
            preceding = self.read_until(["directive"], False)

            # With an immediately-appearing directive, handle its details.

            if preceding == "":
                handler = self.handlers.get(self.matching_pattern())
                if handler:
                    handler(self, region)
                else:
                    break

            # Otherwise, with no immediate directive (or none at all), stop.

            else:
                break

    # Parsing utilities.

    def parse_region_details(self, region, pattern_names, strict=False):

        """
        Search 'region' using the 'pattern_names'. If 'strict' is set to a true
        value, forbid the accumulation of additional textual padding.
        """

        try:
            while True:

                # Obtain text before any marker or the end of the input.

                preceding = self.read_until(pattern_names)
                if preceding:
                    if not strict:
                        region.append_inline(Text(preceding))
                    else:
                        break

                # End of input.

                if not self.matching_pattern():
                    break

                # Obtain any feature.

                feature = self.match_group("feature") or self.match_group()
                handler = self.handlers.get(self.matching_pattern())

                # Handle each feature or add text to the region.

                if handler:
                    handler(self, region)
                elif not strict:
                    region.append_inline(Text(feature))
                else:
                    break

        except StopIteration:
            pass

        region.normalise()

    def add_node(self, region, node):

        "Add to 'region' the given 'node'."

        region.add(node)

    def append_node(self, region, node):

        "Append to 'region' the given 'node'."

        region.append(node)

    def end_region(self, region):

        "End the parsing of 'region', breaking out of the parsing loop."

        raise StopIteration

    def queue_match(self):

        "Queue the current match."

        self.items.queue_match()

    def new_block(self, region):

        "Start a new block in 'region'."

        self.add_node(region, Block([]))

    # Common handler methods.

    def parse_region_end(self, node):

        "Handle the end of a region occurring within 'node'."

        level = self.match_group("level")
        feature = self.match_group("feature")
        self.region.extra = self.match_group("extra")

        if self.region.have_end(level):
            raise StopIteration
        else:
            node.append_inline(Text(feature))

# vim: tabstop=4 expandtab shiftwidth=4
